unit CMDateUtil;

{**********************************************************************
 Datumsroutinen fr Session "DL 17 Y2K: Datumsarithmetik mit Delphi"
 Entwicklerkonferenz 1999
 Compiler: DELPHI 4
 ----------------------------------------------------------------------
 date routines for session "DL 17 Y2K: Date Arithmetic with Delphi"
 Developer Conference 1999
 Compiler: DELPHI 4
 ----------------------------------------------------------------------

 (c) 1999 by Christian Maas
             www.chmaas.handshake.de
             chmaas@hit.sb.sub.de

 **********************************************************************}

interface
uses Windows;

type
  TDateOrder = (doUndef, doMDY, doDMY, doYMD);

{**********************************************************************
 aus Datumsstring YYYYMMDD Jahr als integer zurckgeben
 ----------------------------------------------------------------------
 extract from date string YYYYMMDD year as integer
 **********************************************************************}
function ExtractYear (const YYYYMMDD: string): integer;

{**********************************************************************
 aus Datumsstring YYYYMMDD Monat als integer zurckgeben
 ----------------------------------------------------------------------
 extract from date string YYYYMMDD month as integer
 **********************************************************************}
function ExtractMonth (const YYYYMMDD: string): integer;

{**********************************************************************
 aus Datumsstring YYYYMMDD Tag als integer zurckgeben
 ----------------------------------------------------------------------
 extract from date string YYYYMMDD day as integer
 **********************************************************************}
function ExtractDay (const YYYYMMDD: string): integer;

{**********************************************************************
 Datumsstring YYYYMMDD in TDateTime umwandeln
 ----------------------------------------------------------------------
 convert date string YYYYMMDD to TDateTime
 **********************************************************************}
function YYYYMMDDtoTDateTime (const YYYYMMDD: string): TDateTime;

{**********************************************************************
 <Date> im TDateTime-Format in Datumsstring YYYYMMDD umwandeln
 ----------------------------------------------------------------------
 convert <Date> in TDateTime format to date string YYYYMMDD
 **********************************************************************}
function TDateTimeToYYYYMMDD (const Date: TDateTime): string;

{**********************************************************************
 Datum <Year>, <Month>, <Day> in String YYYYMMDD umwandeln
 ----------------------------------------------------------------------
 convert date given by <Year>, <Month>, <Day> to date string YYYYMMDD
 **********************************************************************}
function YMDtoYYYYMMDD (Year, Month, Day: integer): string;

{**********************************************************************
 aktuelles Datum im Format YYYYMMDD ermitteln
 ----------------------------------------------------------------------
 get current date in format YYYYMMDD
 **********************************************************************}
function GetCurrentDate: string;

{**********************************************************************
 prfen, ob <Year> Schaltjahr ist
 Rckgabewert 0 <=> <Year> ist kein Schaltjahr
              1 <=> <Year> ist Schaltjahr
 ----------------------------------------------------------------------
 determine if <Year> is leap year
 Result       0 <=> <Year> is not leap year
              1 <=> <Year> is leap year
 **********************************************************************}
function IsLeapYear (const Year: integer): integer;

{**********************************************************************
 Anzahl der Tage im Jahr bis zum Ende des angegebenen Monats
 gltige Werte fr <Month>: 0..12
 ----------------------------------------------------------------------
 number of days in year until end of indicated month
 valid values for <Month>: 0..12
 **********************************************************************}
function DaysTillEndOfMonth (const Year, Month: integer): integer;

{**********************************************************************
 Anzahl der Tage im Monat des angegebenen Jahres
 ----------------------------------------------------------------------
 number of days in month of indicated year
 **********************************************************************}
function DaysOfMonth (const Year, Month: integer): integer;

{**********************************************************************
 aus Datum fortlaufende Nummer des Tages im Jahr bestimmen (1..365/366)
 ----------------------------------------------------------------------
 compute continuous number of day within year for YYYYMMDD (1..365/366)
 **********************************************************************}
function DayNumber (const YYYYMMDD: string): integer;

{**********************************************************************
 aus fortlaufender Nummer <n> eines Tages im Jahr Monat (1..12)
 bestimmen, in dem der Tag liegt
 ----------------------------------------------------------------------
 determine month (1..12) from continuous number <n> within year
 **********************************************************************}
function DayNumberToMonth (Year, n: integer): integer;

{**********************************************************************
 aus fortlaufender Nummer <n> eines Tages im Jahr Datum bestimmen
 ----------------------------------------------------------------------
 determine date from continuous number <n> within year
 **********************************************************************}
function DayNumberToYYYYMMDD (Year, n: integer): string;

{**********************************************************************
 Datum des letzten Tages des Monats im angegebenen Jahr
 ----------------------------------------------------------------------
 date of last day of the month within indicated year
 **********************************************************************}
function Ultimo (const Year, Month: integer): string;

{**********************************************************************
 Julianische Tagesnummer (Julian Day Number) eines Datums im
 Gregorianischen Kalender bestimmen (d.h. ab 15.10.1582)
 ----------------------------------------------------------------------
 Julian Day Number of date within Gregorian Calendar (i.e. since
 October 15, 1582)
 **********************************************************************}
function JulianDayNumber (const YYYYMMDD: string): longint;

{**********************************************************************
 Umkehrfunktion zur Julianischen Tagesnummer
 ----------------------------------------------------------------------
 inverse function to Julian Day Number
 **********************************************************************}
function JDNtoYYYYMMDD (n: integer): string;

{**********************************************************************
 ist YYYYMMDD ein gltiges Datum?
 ----------------------------------------------------------------------
 is YYYYMMDD a valid date?
 **********************************************************************}
function IsValidDate (const YYYYMMDD: string): boolean;

{**********************************************************************
 n Tage zum angegebenen Datum addieren
 ----------------------------------------------------------------------
 add n days to indicated date
 **********************************************************************}
function AddDays (const YYYYMMDD: string; n: integer): string;

{**********************************************************************
 Wochentag nach ISO/DIN (Montag=1, Dienstag=2, ..., Sonntag=7)
 ----------------------------------------------------------------------
 week day according to ISO (Monday=1, Tuesday=2, ..., Sunday=7)
 **********************************************************************}
function WeekDayISO (const YYYYMMDD: string): integer;

{**********************************************************************
 kurzer Name des Wochentags (1=Mo, 2=Di, ..., 7=So)
 ----------------------------------------------------------------------
 short name of day (1=Mo, 2=Tu, ..., 7=Su)
 **********************************************************************}
function ShortDayName (i: integer): string;

{**********************************************************************
 langer Name des Wochentags (1=Montag, 2=Dienstag, ..., 7=Sonntag)
 ----------------------------------------------------------------------
 long name of day (1=Monday, 2=Tuesday, ..., 7=Sunday)
 **********************************************************************}
function LongDayName (i: integer): string;

{**********************************************************************
 Datum des n-ten Tages (n fr Montag=1, Dienstag=2, ..., Sonntag=7) in
 der gleichen Woche wie bergebenes Datum
 ----------------------------------------------------------------------
 date of n-th day (n for Monday=1, Tuesday=2, ..., Sunday=7) within
 same week as indicated date
 **********************************************************************}
function NthDayOfSameWeek (const YYYYMMDD: string; const n: integer): string;

{**********************************************************************
 Datum des Beginns der 1. Kalenderwoche im Jahr nach ISO/DIN
 ----------------------------------------------------------------------
 date of begin of first week in year according to ISO
 **********************************************************************}
function BeginWeek1ISO (Year: integer): string;

{**********************************************************************
 Nummer der letzten Kalenderwoche im Jahr nach ISO/DIN
 ----------------------------------------------------------------------
 number of last week in year according to ISO
 **********************************************************************}
function NumberLastWeekISO (Year: integer): integer;

{**********************************************************************
 Kalenderwoche nach ISO/DIN eines Datums; zustzlich wird das Jahr, in
 das die Woche fllt, zurckgegeben
 ----------------------------------------------------------------------
 number of week of indicated date; also returned: year belonging to week
 **********************************************************************}
function WeekNumberISO (const YYYYMMDD: string; var Year: integer): integer;

{**********************************************************************
 Ostersonntag im Jahr
 ----------------------------------------------------------------------
 Easter Sunday in year
 **********************************************************************}
function EasterSunday (Year: integer): string;

{**********************************************************************
 Aufruf der Win98-API-Funktion GetCalendarInfoA (wenn verfgbar)
 ----------------------------------------------------------------------
 call Win98-API-function GetCalendarInfoA if available
 **********************************************************************}
function GetCalendarInfo (Locale: LCID; Calendar: CALID;
                          CalendType: CALTYPE; lpCalData : lpStr;
                          cchData: Integer; lpValue: lpDWord): integer;

{**********************************************************************
 Pivot-Jahr (max. Wert fr Interpretation eines zweistelligen Jahres)
 anhand Einstellung in Systemsteuerung bestimmen
 ----------------------------------------------------------------------
 get pivot year (maximum value for 2-digit-year) from Control Panel
 settings
 **********************************************************************}
function GetPivotYear: integer;

{**********************************************************************
 Jahr innerhalb des durch Pivot-Jahr definierten Fensters?
 ----------------------------------------------------------------------
 year within frame defined by pivot year?
 **********************************************************************}
function WithinPivotFrame (Year: integer): boolean;

{**********************************************************************
 zweistellige Jahreszahl anhand Pivot-Jahr vierstellig interpretieren
 ----------------------------------------------------------------------
 interpret 2-digit-year according to pivot year with 4 digits
 **********************************************************************}
function YYtoYYYY (YY: integer): integer;

{**********************************************************************
 ShortDateFormat mit 4-stelligem Jahr erzeugen
 ----------------------------------------------------------------------
 generate ShortDateFormat with 4 digits for year
 **********************************************************************}
function ShortDateFormatYYYY: string;

{**********************************************************************
 Reihenfolge von Jahr/Monat/Tag im Datumsformat <DateFormat>;
 es gibt nur 3 gltige Reihenfolgen:
    - Jahr/Monat/Tag (ISO)
    - Monat/Tag/Jahr (USA)
    - Tag/Monat/Jahr (Europa)
 ----------------------------------------------------------------------
 order of Year/month/day in date format <DateFormat>; there are only
 three valid orders:
    - Year/Month/Day (ISO)
    - Month/Day/Year (USA)
    - Day/Month/Year (Europe)
 **********************************************************************}
function GetDateOrder (DateFormat: string): TDateOrder;

{**********************************************************************
 Vom Benutzer eingegebenes Datum <DateString> auf Gltigkeit prfen.
 Rckgabewert der Funktion = TRUE bei gltigem Datum; in diesem Fall
 enthlt YYYYMMDD das Datum im ISO-Format
 ----------------------------------------------------------------------
 Validate date <DateString> entered by user. Result = TRUE if valid;
 in this case, YYYYMMDD contains ISO date
 **********************************************************************}
function ScanDateString (DateString: string; var YYYYMMDD: string): boolean;

{**********************************************************************
 Datum entsprechend ShortDateFormat mit 2- oder 4-stelligem Jahr
 formatieren
 ----------------------------------------------------------------------
 format date according to ShortDateFormat with 2 or 4 digits for year
 **********************************************************************}
function FormatDate (const YYYYMMDD: string; const Year4Digits: boolean): string;

{**********************************************************************
 Pivot-Jahr (max. Wert fr Interpretation eines zweistelligen Jahres)
 gem Einstellung in Systemsteuerung
 ----------------------------------------------------------------------
 pivot year (maximum value for 2-digit-year) according to Control Panel
 settings
 **********************************************************************}
const
  PivotYear: integer = 2029;

implementation
uses SysUtils;

const
  DateCorrection = 1721425;
  BeginGregorianCal = '15821015';

{**********************************************************************
 aus Datumsstring YYYYMMDD Jahr als integer zurckgeben
 ----------------------------------------------------------------------
 extract from date string YYYYMMDD year as integer
 **********************************************************************}
function ExtractYear (const YYYYMMDD: string): integer;
begin
  try
    Result := StrToInt (copy (YYYYMMDD, 1, 4));
  except
    Result := 0;
  end;
end;

{**********************************************************************
 aus Datumsstring YYYYMMDD Monat als integer zurckgeben
 ----------------------------------------------------------------------
 extract from date string YYYYMMDD month as integer
 **********************************************************************}
function ExtractMonth (const YYYYMMDD: string): integer;
begin
  try
    Result := StrToInt (copy (YYYYMMDD, 5, 2));
  except
    Result := 0;
  end;
end;

{**********************************************************************
 aus Datumsstring YYYYMMDD Tag als integer zurckgeben
 ----------------------------------------------------------------------
 extract from date string YYYYMMDD day as integer
 **********************************************************************}
function ExtractDay (const YYYYMMDD: string): integer;
begin
  try
    Result := StrToInt (copy (YYYYMMDD, 7, 2));
  except
    Result := 0;
  end;
end;

{**********************************************************************
 Datumsstring YYYYMMDD in TDateTime umwandeln
 ----------------------------------------------------------------------
 convert date string YYYYMMDD to TDateTime
 **********************************************************************}
function YYYYMMDDtoTDateTime (const YYYYMMDD: string): TDateTime;
begin
  try
    Result := EncodeDate (ExtractYear (YYYYMMDD), ExtractMonth (YYYYMMDD),
      ExtractDay (YYYYMMDD));
  except
    Result := 0;
  end;
end;

{**********************************************************************
 <Date> im TDateTime-Format in Datumsstring YYYYMMDD umwandeln
 ----------------------------------------------------------------------
 convert <Date> in TDateTime format to date string YYYYMMDD
 **********************************************************************}
function TDateTimeToYYYYMMDD (const Date: TDateTime): string;
var Year, Month, Day: word;
begin
  try
    DecodeDate (Date, Year, Month, Day);
    Result := Format ('%4.4d%2.2d%2.2d', [Year, Month, Day]);
  except
    Result := '       ';
  end;
end;

{**********************************************************************
 Datum <Year>, <Month>, <Day> in String YYYYMMDD umwandeln
 ----------------------------------------------------------------------
 convert date given by <Year>, <Month>, <Day> to date string YYYYMMDD
 **********************************************************************}
function YMDtoYYYYMMDD (Year, Month, Day: integer): string;
begin
  Result := Format ('%4.4d%2.2d%2.2d', [Year, Month, Day]);
end;

{**********************************************************************
 aktuelles Datum im Format YYYYMMDD ermitteln
 ----------------------------------------------------------------------
 get current date in format YYYYMMDD
 **********************************************************************}
function GetCurrentDate: string;
var SystemTime: TSystemTime;
begin
  GetLocalTime (SystemTime);
  with SystemTime do
    Result := YMDtoYYYYMMDD (wYear, wMonth, wDay);
end;

{**********************************************************************
 prfen, ob <Year> Schaltjahr ist
 Rckgabewert 0 <=> <Year> ist kein Schaltjahr
              1 <=> <Year> ist Schaltjahr
 ----------------------------------------------------------------------
 determine if <Year> is leap year
 Result       0 <=> <Year> is not leap year
              1 <=> <Year> is leap year
 **********************************************************************}
function IsLeapYear (const Year: integer): integer;
begin
  if Year mod 400 = 0 then
    { Ausnahme von der Ausnahme: durch 400 teilbare Jahr sind Schaltjahre }
    Result := 1
  else if Year mod 100 = 0 then
    { Ausnahme von der Regel: durch 100 teilbare Jahr sind keine Schaltjahre }
    Result := 0
  else if Year mod 4 = 0 then
    { Regel: durch 4 teilbare Jahr sind Schaltjahre }
    Result := 1
  else
    Result := 0;
end;

{**********************************************************************
 Anzahl der Tage im Jahr bis zum Ende des angegebenen Monats
 gltige Werte fr <Month>: 0..12
 ----------------------------------------------------------------------
 number of days in year until end of indicated month
 valid values for <Month>: 0..12
 **********************************************************************}
function DaysTillEndOfMonth (const Year, Month: integer): integer;
begin
  if Month < 2 then
    Result := Month * 31
  else
    Result := (153 * Month - 9) div 5 + IsLeapYear (Year);
end;

{**********************************************************************
 Anzahl der Tage im Monat des angegebenen Jahres
 ----------------------------------------------------------------------
 number of days in month of indicated year
 **********************************************************************}
function DaysOfMonth (const Year, Month: integer): integer;
begin
  Result :=
    { Tage bis zum Ende des Monats / days until end of month }
    DaysTillEndOfMonth (Year, Month) -
    { Tage bis zum Ende des Vormonats / days until end of previous month }
    DaysTillEndOfMonth (Year, Month - 1);
end;

{**********************************************************************
 aus Datum fortlaufende Nummer des Tages im Jahr bestimmen (1..365/366)
 ----------------------------------------------------------------------
 compute continuous number of day within year for YYYYMMDD (1..365/366)
 **********************************************************************}
function DayNumber (const YYYYMMDD: string): integer;
begin
  DayNumber :=
    { Tage bis Ende Vormonat / day until end of previous month }
    DaysTillEndOfMonth (ExtractYear (YYYYMMDD), ExtractMonth (YYYYMMDD) - 1) +
    { plus Tag / add day}
    ExtractDay (YYYYMMDD);
end;

{**********************************************************************
 aus fortlaufender Nummer <n> eines Tages im Jahr Monat (1..12)
 bestimmen, in dem der Tag liegt
 ----------------------------------------------------------------------
 determine month (1..12) from continuous number <n> within year
 **********************************************************************}
function DayNumberToMonth (Year, n: integer): integer;
begin
  if n <= 31 then
    Result := 1
  else
    Result := (5 * n + 161 - 5 * IsLeapYear (Year)) div 153;
end;

{**********************************************************************
 aus fortlaufender Nummer <n> eines Tages im Jahr Datum bestimmen
 ----------------------------------------------------------------------
 determine date from continuous number <n> within year
 **********************************************************************}
function DayNumberToYYYYMMDD (Year, n: integer): string;
var Month, Day: integer;
begin
  Month  := DayNumberToMonth (Year, n);
  Day    := n - DaysTillEndOfMonth (Year, Month - 1);
  Result := YMDtoYYYYMMDD (Year, Month, Day)
end;

{**********************************************************************
 Datum des letzten Tages des Monats im angegebenen Jahr
 ----------------------------------------------------------------------
 date of last day of the month within indicated year
 **********************************************************************}
function Ultimo (const Year, Month: integer): string;
begin
  Result := DayNumberToYYYYMMDD (Year, DaysTillEndOfMonth (Year, Month));
end;

{**********************************************************************
 Anzahl der Tage bis zum Ende des angegebenen Jahres ermitteln.
 ACHTUNG: beginnend bei 01.01.0001 unter Annahme der durchgehenden
 Gltigkeit des Gregorianischen Kalenders! (nur intern benutzt)
 ----------------------------------------------------------------------
 number of days until end of indicated year
 NOTE: beginning with January 1, 0001 supposed Gregorian Calendar
 were already in use! (used only internally)
 **********************************************************************}
function DaysTillEndOfYear (const Jahr: longint): longint;
begin
  Result := Jahr * 365 + Jahr div 4 - Jahr div 100 + Jahr div 400;
end;

{**********************************************************************
 Julianische Tagesnummer (Julian Day Number) eines Datums im
 Gregorianischen Kalender bestimmen (d.h. ab 15.10.1582)
 ----------------------------------------------------------------------
 Julian Day Number of date within Gregorian Calendar (i.e. since
 October 15, 1582)
 **********************************************************************}
function JulianDayNumber (const YYYYMMDD: string): longint;
begin
  Result :=
    { Tage bis Ende Vorjahr / days until end of prevoious year }
    DaysTillEndOfYear (ExtractYear (YYYYMMDD) - 1) +
    { Tag im Jahr / continuous number of day within year }
    longint (DayNumber (YYYYMMDD)) +
    { Korrektur / correction needed }
    DateCorrection;
end;

{**********************************************************************
 Umkehrfunktion zur Julianischen Tagesnummer
 ----------------------------------------------------------------------
 inverse function to Julian Day Number
 **********************************************************************}
function JDNtoYYYYMMDD (n: integer): string;
var Year: longint;
begin
  { bestimme Jahr zu klein / compute Year too small }
  n := n - DateCorrection;
  Year := n div 365;

  { bestimme kleinste obere Schranke fr Jahr / compute Year too large }
  repeat
    inc (Year)
  until DaysTillEndOfYear (Year - 1) > n;

  { bestimme grte untere Schranke fr Jahr / compute Year too small }
  repeat
    dec (Year);
  until DaysTillEndOfYear (Year - 1) < n;

  Result := DayNumberToYYYYMMDD (Year, n - DaysTillEndOfYear (Year - 1));

end;

{**********************************************************************
 ist YYYYMMDD ein gltiges Datum?
 ----------------------------------------------------------------------
 is YYYYMMDD a valid date?
 **********************************************************************}
function IsValidDate (const YYYYMMDD: string): boolean;
begin
  { wir nutzen aus, da die Zuordnung eines Datums zur Julianischen
    Tagesnummer eineindeutig ist (Bijektion): fr ein ungltiges Datum
    wird zwar irgend eine Zahl n als JDN zurckgegeben, das fr n
    rckermittelte Datum ist jedoch gltig und damit zwangslufig
    nicht identisch mit dem ursprnglichen ungltigen Datum
    --------------------------------------------------------------------
    we use the fact that the relation between dates and corresponding
    Julian Day Numbers is a bijection:
    for an invalid date some integer n is computed as JDN, but the date
    computed for n is valid and therefore not equal to the original
    invalid date }
  Result := (YYYYMMDD = JDNtoYYYYMMDD (JulianDayNumber (YYYYMMDD))) and
            (YYYYMMDD >= BeginGregorianCal);
end;

{**********************************************************************
 n Tage zum angegebenen Datum addieren
 ----------------------------------------------------------------------
 add n days to indicated date
 **********************************************************************}
function AddDays (const YYYYMMDD: string; n: integer): string;
begin
  Result := JDNtoYYYYMMDD (JulianDayNumber (YYYYMMDD) + n);
end;

{**********************************************************************
 Wochentag nach ISO/DIN (Montag=1, Dienstag=2, ..., Sonntag=7)
 ----------------------------------------------------------------------
 week day according to ISO (Monday=1, Tuesday=2, ..., Sunday=7)
 **********************************************************************}
function WeekDayISO (const YYYYMMDD: string): integer;
begin
  Result := JulianDayNumber (YYYYMMDD) mod 7 + 1;
end;

{**********************************************************************
 kurzer Name des Wochentags (1=Mo, 2=Di, ..., 7=So)
 ----------------------------------------------------------------------
 short name of day (1=Mo, 2=Tu, ..., 7=Su)
 **********************************************************************}
function ShortDayName (i: integer): string;
begin
  { Mo Tu We Th Fr Sa Su
     1  2  3  4  5  6  7  ISO/DIN
     2  3  4  5  6  7  1  ShortDayNames in DELPHI }
  Result := ShortDayNames [(i mod 7) + 1];
end;

{**********************************************************************
 langer Name des Wochentags (1=Montag, 2=Dienstag, ..., 7=Sonntag)
 ----------------------------------------------------------------------
 long name of day (1=Monday, 2=Tuesday, ..., 7=Sunday)
 **********************************************************************}
function LongDayName (i: integer): string;
begin
  { Mo Tu We Th Fr Sa Su
     1  2  3  4  5  6  7  ISO/DIN
     2  3  4  5  6  7  1  LongDayNames in DELPHI }
  Result := LongDayNames [(i mod 7) + 1];
end;

{**********************************************************************
 Datum des n-ten Tages (n fr Montag=1, Dienstag=2, ..., Sonntag=7) in
 der gleichen Woche wie bergebenes Datum
 ----------------------------------------------------------------------
 date of n-th day (n for Monday=1, Tuesday=2, ..., Sunday=7) within
 same week as indicated date
 **********************************************************************}
function NthDayOfSameWeek (const YYYYMMDD: string; const n: integer): string;
begin
  Result :=
    JDNtoYYYYMMDD (JulianDayNumber (YYYYMMDD) - WeekDayISO (YYYYMMDD) + n);
    {              <------------------------------------------------>
                        vorhergehender Sonntag / previous Sunday           }
end;

{**********************************************************************
 Datum des Beginns der 1. Kalenderwoche im Jahr nach ISO/DIN
 ----------------------------------------------------------------------
 date of begin of first week in year according to ISO
 **********************************************************************}
function BeginWeek1ISO (Year: integer): string;
var January4th: string;
begin
  { 4. Januar liegt stets in 1. KW / January 4 always within 1st week }
  January4th := YMDtoYYYYMMDD (Year, 1, 4);
  { Montag vor 4. Januar ist Beginn der 1. KW /
    Monday before January 4 is begin of 1st week }
  Result := NthDayOfSameWeek (January4th, 1);
end;

{**********************************************************************
 Nummer der letzten Kalenderwoche im Jahr nach ISO/DIN
 ----------------------------------------------------------------------
 number of last week in year according to ISO
 **********************************************************************}
function NumberLastWeekISO (Year: integer): integer;
begin
  Result := (
    { Beginn der 1. KW des Folgejahres / begin of 1st week of next year }
    JulianDayNumber (BeginWeek1ISO (Year + 1)) -
    { Beginn der ersten Woche des Jahres / begin of first week of year }
    JulianDayNumber (BeginWeek1ISO (Year)))
    div 7;
end;

{**********************************************************************
 Kalenderwoche nach ISO/DIN eines Datums; zustzlich wird das Jahr, in
 das die Woche fllt, zurckgegeben
 ----------------------------------------------------------------------
 number of week of indicated date; also returned: year belonging to week
 **********************************************************************}
function WeekNumberISO (const YYYYMMDD: string; var Year: integer): integer;
var MondayWeek1: string;
begin

  (*  1. Mglichkeit / 1st way

  { Jahr der KW wird durch den Donnerstag der Woche bestimmt /
    year of week is determined by Thursday of same week }
  Year := ExtractYear (NthDayOfSameWeek (YYYYMMDD, 4));

  Result := (
    { JDN des vorangehenden Montags / JDN of previous Monday }
    JulianDayNumber (NthDayOfSameWeek (YYYYMMDD, 1)) -
    { 4. Januar des Jahres der KW / January 4 of year of week }
    JulianDayNumber (YMDtoYYYYMMDD (Year, 1, 4)) + 13)
    div 7;*)

   { 2. Mglichkeit / 2nd way }

   Year := ExtractYear (YYYYMMDD);
   if YYYYMMDD >= BeginWeek1ISO (Year + 1) then begin
     { Datum in 1. KW des Folgejahres / date within 1st week of next year }
     inc (Year);
     Result := 1;
   end
   else begin
     MondayWeek1 := BeginWeek1ISO (Year);
     if YYYYMMDD < MondayWeek1 then begin
       { Datum in letzter KW des Vorjahres / date within last week of previous year }
       dec (Year);
       Result := NumberLastWeekISO (Year);
     end
     else
       Result := (
         JulianDayNumber (NthDayOfSameWeek (YYYYMMDD, 1)) -
         JulianDayNumber (MondayWeek1))
         div 7 + 1;
   end;

end;

{**********************************************************************
 Ostersonntag im Jahr (Algorithmus von Oudin, 1940)
 ----------------------------------------------------------------------
 Easter Sunday in year (algorithm of Oudin, 1940)
 **********************************************************************}
function EasterSunday (Year: integer): string;
var G: integer; // Golden Number - 1
    C: integer; // Century - 1
    H: integer; // (23 - Epact) modulo 30
    I: integer; // number of days from 21 March to the Paschal full moon
    J: integer; // the weekday for the Paschal full moon (0=Sunday, 1=Monday...
    L: integer; // number of days from 21 March to the Sunday on or before
                // the Paschal full moon (a number between -6 and 28)
    EasterMonth: integer;
    EasterDay: integer;
begin
  if Year < 1583 then
    Result := ''
  else begin
    G := Year mod 19;
    C := Year div 100;
    H := (C - C div 4 - (8 * C + 13) div 25 + 19 * G + 15) mod 30;
    I := H - (H div 28) * (1 - (H div 28) * (29 div (H + 1)) * ((21 - G) div 11));
    J := (Year + Year div 4 + I + 2 - C + C div 4) mod 7;
    L := I - J;
    EasterMonth := 3 + (L + 40) div 44;
    EasterDay   := L + 28 - 31 * (EasterMonth div 4);
    Result := YMDtoYYYYMMDD (Year, EasterMonth, EasterDay);
  end;
end;

{**********************************************************************
 Aufruf der Win98-API-Funktion GetCalendarInfoA (wenn verfgbar)
 ----------------------------------------------------------------------
 call Win98-API-function GetCalendarInfoA if available
 **********************************************************************}
function GetCalendarInfo (Locale: LCID; Calendar: CALID; CalendType: CALTYPE;
  lpCalData : lpStr; cchData: integer; lpValue: lpDWord): integer;
type
  TGetCalendarInfo =
    function (Locale: LCID; Calendar: CALID; CalendType: CALTYPE;
              lpCalData: lpStr; cchData: Integer;
              lpValue: lpDWord): integer; stdcall;
var
  hKernelDLL : Integer;
  MyGetCalendarInfo: TGetCalendarInfo;

begin
  Result:= 0;
  hKernelDLL:= LoadLibrary('Kernel32');
  if hKernelDLL <> 0 then
    try
      if GetProcAddress (hKernelDll, 'GetCalendarInfoA') <> nil then begin
        MyGetCalendarInfo := GetProcAddress (hKernelDll, 'GetCalendarInfoA');
        Result := MyGetCalendarInfo (Locale, Calendar, CalendType, lpCalData,
                                     cchData, lpValue);
      end;
    finally
      FreeLibrary (hKernelDLL);
    end;
end;

{**********************************************************************
 Pivot-Jahr (max. Wert fr Interpretation eines zweistelligen Jahres)
 anhand Einstellung in Systemsteuerung bestimmen
 ----------------------------------------------------------------------
 get pivot year (maximum value for 2-digit-year) from Control Panel
 **********************************************************************}
function GetPivotYear: integer;
var Buffer: array [0..16] of char;
begin
  Result := 2029;
  { now retrieve W98 or NT 5 ITwoDigitYearMax value }
  if GetCalendarInfo (GetUserDefaultLCID, CAL_GREGORIAN, $30, Buffer,
    SizeOf (Buffer), nil) > 0
  then
    Result := StrToInt (string (Buffer));
end;

{**********************************************************************
 Jahr innerhalb des durch Pivot-Jahr definierten Fensters?
 ----------------------------------------------------------------------
 year within frame defined by pivot year?
 **********************************************************************}
function WithinPivotFrame (Year: integer): boolean;
begin
  Result := (Year >= PivotYear - 99) and (Year <= PivotYear);
end;

{**********************************************************************
 zweistellige Jahreszahl anhand Pivot-Jahr vierstellig interpretieren
 ----------------------------------------------------------------------
 interpret 2-digit-year according to pivot year with 4 digits
 **********************************************************************}
function YYtoYYYY (YY: integer): integer;
var k: integer;
begin
  { YYYY <= PivotYear gefordert; maximales k gesucht, so da /
    YYYY <= PivotYear needed; determine maximum value k with

    YY + k * 100 <= PivotYear      <=>
    k <= (PivotYear - YY) div 100      }
  k := (PivotYear - YY) div 100;
  Result := YY + k * 100;
  if Result > PivotYear then
    dec (Result, 100);
end;

{**********************************************************************
 letztes Vorkommen von Substring (nur intern benutzt)
 ----------------------------------------------------------------------
 last occurence of substring (used only internally)
 **********************************************************************}
function LastPos (Substr: string; s: string): integer;
var i: integer;
begin
  if length (s) < length (Substr) then
    Result := 0
  else begin
    for i := length (s) - length (Substr) + 1 downto 1 do
      if copy (s, i, length (Substr)) = Substr then begin
        Result := i;
        Exit;
      end;
    Result := 0;
  end;
end;

{**********************************************************************
 ShortDateFormat mit 4-stelligem Jahr erzeugen
 ----------------------------------------------------------------------
 generate ShortDateFormat with 4 digits for year
 **********************************************************************}
function ShortDateFormatYYYY: string;
var FPos, LPos: integer;
begin
  Result := UpperCase (ShortDateFormat);
  FPos := Pos ('Y', Result);
  LPos := LastPos ('Y', Result);
  if FPos = 0 then
    { Jahr nicht angegeben / year not indicated }
    Result := Result + DateSeparator + 'YYYY'
  else if LPos - FPos < 3 then
    { durch 4-stelliges Jahr ersetzen / replace by 4-digit year }
    Result := copy (Result, 1, FPos - 1) + 'YYYY' +
      copy (Result, LPos + 1, Length (Result) - LPos);
end;

{**********************************************************************
 Reihenfolge von Jahr/Monat/Tag im Datumsformat <DateFormat> anhand des
 ersten signifikanten Zeichens; es gibt nur 3 gltige Reihenfolgen:
    - Jahr/Monat/Tag (ISO)
    - Monat/Tag/Jahr (USA)
    - Tag/Monat/Jahr (Europa)
 ----------------------------------------------------------------------
 order of Year/month/day in date format <DateFormat> according to first
 significant character; there are only three valid orders:
    - Year/Month/Day (ISO)
    - Month/Day/Year (USA)
    - Day/Month/Year (Europe)
 **********************************************************************}
function GetDateOrder (DateFormat: string): TDateOrder;
var i: integer;
begin
  Result := doUndef;
  for i := 1 to length (DateFormat) do begin
    case DateFormat [i] of
      'Y', 'y': begin Result := doYMD; Exit; end;
      'M', 'm': begin Result := doMDY; Exit; end;
      'D', 'd': begin Result := doDMY; Exit; end;
    end;
  end;
end;

{**********************************************************************
 Vom Benutzer eingegebenes Datum <DateString> auf Gltigkeit prfen.
 Rckgabewert der Funktion = TRUE bei gltigem Datum; in diesem Fall
 enthlt YYYYMMDD das Datum im ISO-Format
 ----------------------------------------------------------------------
 Validate date <DateString> entered by user. Result = TRUE if valid;
 in this case, YYYYMMDD contains ISO date
 **********************************************************************}
function ScanDateString (DateString: string; var YYYYMMDD: string): boolean;

var DateOrder: TDateOrder;
    iSeparator, iLen: integer;
    Part: string;
    YearStr, MonthStr, DayStr: string;
    Year, Month, Day: word;
    JDN: longint;

begin

  YYYYMMDD := '';  { Rckgabewert fr gltiges Datum }
  Result := false; { Datum noch nicht verifiziert }

  if pos (DateSeparator + DateSeparator, DateString) > 0 then
    { mehrfaches Vorkommen des Datumsseparators hintereinander }
    Exit;

  DateOrder := GetDateOrder (ShortDateFormat);
  if DateOrder = doUndef then
    Exit;

  DateString := Trim (DateString);
  YearStr := '';
  MonthStr := '';
  DayStr := '';

  { Analyse der 1. Datumskomponente / scan 1st date component }
  iSeparator := pos (DateSeparator, DateString);
  if iSeparator = 0 then begin
    { Datum ohne Trenner / date without separator }
    if (DateOrder = doYMD) and (length (DateString) = 8) then
      { Jahr 4-stellig angegeben / year indicated with 4 digits }
      iLen := 4
    else
      iLen := 2;
    Part := copy (DateString, 1, iLen);
    DateString := copy (DateString, iLen + 1, length (DateString) - iLen);
  end
  else begin
    Part := copy (DateString, 1, iSeparator - 1);
    DateString := copy (DateString, iSeparator + 1, length (DateString) - iSeparator);
  end;

  case DateOrder of
    doMDY: MonthStr := Part;
    doDMY: DayStr   := Part;
    doYMD: YearStr  := Part;
  end;

  { Analyse der 2. Datumskomponente / scan 2nd date component }
  iSeparator := pos (DateSeparator, DateString);
  if iSeparator = 0 then begin
    Part := copy (DateString, 1, 2);
    DateString := copy (DateString, 3, length (DateString) - 2);
  end
  else begin
    Part := copy (DateString, 1, iSeparator - 1);
    DateString := copy (DateString, iSeparator + 1, length (DateString) - iSeparator);
  end;

  case DateOrder of
    doMDY: DayStr   := Part;
    doDMY: MonthStr := Part;
    doYMD: MonthStr := Part;
  end;

  { Analyse der als Rest verbleibenden 3. Datumskomponente /
    scan remaining 3rd date component }
  case DateOrder of
    doMDY: YearStr  := DateString;
    doDMY: YearStr  := DateString;
    doYMD: DayStr   := DateString;
  end;

  { Monat und/oder Jahr ggfs. durch aktuelle Angaben ersetzen /
    substitute month and/or year by current values if needed }
  try
    DecodeDate (Date, Year, Month, Day);
    if MonthStr = '' then
      MonthStr := IntToStr (Month);
    if YearStr = '' then
      YearStr := IntToStr (Year);
    Year  := StrToInt (YearStr);
    Month := StrToInt (MonthStr);
    Day   := StrToInt (DayStr);
  except
    Exit;
  end;

  if Year < 100 then
    Year := YYtoYYYY (Year);

  YYYYMMDD := Format ('%4.4d%2.2d%2.2d', [Year, Month, Day]);
  JDN := JulianDayNumber (YYYYMMDD);
  if (JDN >= 2299161) and (JDNtoYYYYMMDD (JDN) = YYYYMMDD) then
    { gltiges Datum innerhalb Gregorianischem Kalender /
      valid date within Gregorian Calendar }
    Result := true
  else
    YYYYMMDD := '';

end;

{**********************************************************************
 Datum entsprechend ShortDateFormat mit 2- oder 4-stelligem Jahr
 formatieren
 ----------------------------------------------------------------------
 format date according to ShortDateFormat with 2- or 4-digit year
 **********************************************************************}
function FormatDate (const YYYYMMDD: string; const Year4Digits: boolean): string;

var DateOrder: TDateOrder;
    FormatStr: string;
    Year, Month, Day: integer;

begin

  if length (YYYYMMDD) < 8 then begin
    { ungltiges datum / invalid date }
    Result := '';
    Exit;
  end;

  DateOrder := GetDateOrder (ShortDateFormat);

  if DateOrder = doUndef then
    { falls undefiniert, ISO-Norm benutzen / if undefined, use ISO standard }
    DateOrder := doYMD;

  Year  := ExtractYear  (YYYYMMDD);
  Month := ExtractMonth (YYYYMMDD);
  Day   := ExtractDay   (YYYYMMDD);

  if not Year4Digits then begin
    { Jahr nur zweistellig / year with only 2 digits }
    Year := Year mod 100;
    FormatStr := '%2.2d' + DateSeparator + '%2.2d' + DateSeparator + '%2.2d';
  end
  else if DateOrder = doYMD then
    FormatStr := '%4.4d' + DateSeparator + '%2.2d' + DateSeparator + '%2.2d'
  else
    FormatStr := '%2.2d' + DateSeparator + '%2.2d' + DateSeparator + '%4.4d';

  case DateOrder of
    doMDY: Result := Format (FormatStr, [Month, Day, Year]);
    doDMY: Result := Format (FormatStr, [Day, Month, Year]);
    doYMD: Result := Format (FormatStr, [Year, Month, Day]);
  end;

end;

initialization
  {**********************************************************************
   Pivot-Jahr (max. Wert fr Interpretation eines zweistelligen Jahres)
   gem Einstellung in Systemsteuerung
   ----------------------------------------------------------------------
   pivot year (maximum value for 2-digit-year) according to Control Panel
   settings
   **********************************************************************}
  PivotYear := GetPivotYear;

end.
