unit Datum_MW;

{**********************************************************************
 Demo-Anwendung fr Session "DL 17 Y2K: Datumsarithmetik mit Delphi"
 Entwicklerkonferenz 1999
 Compiler: DELPHI 4
 ----------------------------------------------------------------------
 demo application for session "DL 17 Y2K: Datumsarithmetik mit Delphi"
 Developer Conference 1999
 Compiler: DELPHI 4
 ----------------------------------------------------------------------

 (c) 1999 by Christian Maas
             www.chmaas.handshake.de
             chmaas@hit.sb.sub.de

 **********************************************************************}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, ExtCtrls, StdCtrls, Mask, UCMDateEdt;

type
  TForm1 = class(TForm)
    Panel1: TPanel;
    PageControl1: TPageControl;
    TabSheet1: TTabSheet;
    TabSheet2: TTabSheet;
    GroupBox1: TGroupBox;
    ButtonDate2DayNumber: TButton;
    ButtonCancelCoeff: TButton;
    ListViewKoeff: TListView;
    ButtonDayNumber2Date: TButton;
    CheckBoxSJ: TCheckBox;
    TabSheet3: TTabSheet;
    GroupBox2: TGroupBox;
    Label1: TLabel;
    EditJDNVon: TEdit;
    Label2: TLabel;
    MaskEditJDNTag: TMaskEdit;
    Label3: TLabel;
    MaskEditJDNMonat: TMaskEdit;
    MaskEditJDNJahr: TMaskEdit;
    Label4: TLabel;
    ButtonStartJDN: TButton;
    TabSheet4: TTabSheet;
    GroupBox3: TGroupBox;
    RadioButton1999: TRadioButton;
    RadioButton2000: TRadioButton;
    ButtonStartDayNumber: TButton;
    ButtonCancelJDN: TButton;
    GroupBox4: TGroupBox;
    Label6: TLabel;
    EditDatum: TEdit;
    ButtonScan: TButton;
    LabelYYYYMMDD: TLabel;
    LabelDatumFormatiert: TLabel;
    GroupBox5: TGroupBox;
    Label7: TLabel;
    ButtonOstern: TButton;
    LabelOstersonntag: TLabel;
    MaskEditOsterjahr: TMaskEdit;
    LabelKW: TLabel;
    GroupBoxTCMDateEdt: TGroupBox;
    ButtonValidate: TButton;
    TabSheet5: TTabSheet;
    ListViewKW: TListView;
    ButtonKW: TButton;
    Label5: TLabel;
    Label8: TLabel;
    MaskEditJahrVon: TMaskEdit;
    MaskEditJahrBis: TMaskEdit;
    LabelYYYYMMDDvalid: TLabel;
    procedure ButtonAbbruchClick(Sender: TObject);
    procedure ButtonDate2DayNumberClick(Sender: TObject);
    procedure ButtonDayNumber2DateClick(Sender: TObject);
    procedure ButtonStartJDNClick(Sender: TObject);
    procedure ButtonStartDayNumberClick(Sender: TObject);
    procedure ButtonScanClick(Sender: TObject);
    procedure ButtonOsternClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure ButtonValidateClick(Sender: TObject);
    procedure ButtonKWClick(Sender: TObject);
  private
    { Private-Deklarationen }
    Cancelled: boolean;
  public
    { Public-Deklarationen }
    CMDateEdt1: TCMDateEdt;
  end;

var
  Form1: TForm1;

implementation
uses CMDateUtil;

{$R *.DFM}

{**********************************************************************
 Konstanten fr Verifikationsalgorithmen
 ----------------------------------------------------------------------
 constants for verification algorithms
 **********************************************************************}

const
  { Summen ber Tage der Monate 0..12 in Nicht-Schaltjahren: Funktion s (m)
    --------------------------------------------------------------------
    sum of days within months 0..12 (function s (m) for non-leap years) }
  s: array [0..12] of integer =
    (0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365);

  { Summen ber Tage der Monate 0..12 in Schaltjahren: Funktion s (m)
    --------------------------------------------------------------------
    sum of days within months 0..12 (function s (m) for leap years) }
  sly: array [0..12] of integer =
    (0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366);

  { fortlaufende Nummer des Tages im Jahr fr Nicht-Schaltjahre, Werte
    fr Monatsgrenzen
    --------------------------------------------------------------------
    continuous number of day within year (non-leap years), values for
    months' borders }
  n: array [0..23] of integer =
    (  1,  31,  32,  59,  60,  90,  91, 120, 121, 151, 152, 181,
     182, 212, 213, 243, 244, 273, 274, 304, 305, 334, 335, 365);

  { fortlaufende Nummer des Tages im Jahr fr Schaltjahre, Werte
    fr Monatsgrenzen
    --------------------------------------------------------------------
    continuous number of day within year (leap years), values for
    months' borders } 
  nly: array [0..23] of integer =
    (  1,  31,  32,  60,  61,  91,  92, 121, 122, 152, 153, 182,
     183, 213, 214, 244, 245, 274, 275, 305, 306, 335, 336, 366);

  { m [i] ist zu n [i] oder nly [i] gehrender Monat
    --------------------------------------------------------------------
    m [i] is month belonging to n [i] or nly [i] }
  m: array [0..23] of integer =
    (  1,   1,   2,   2,   3,   3,   4,   4,   5,   5,   6,  6,
       7,   7,   8,   8,   9,   9,  10,  10,  11,  11,  12, 12);

{**********************************************************************
 in beiden Algorithmen zur Ermittlung der Koeffizienten verwendete
 Funktion
 ----------------------------------------------------------------------
 function used in both algorithms to compute coefficients
 **********************************************************************}
function f (a, b, c, x: integer): integer;
begin
  Result := (a * x + b) div c;
end;

{**********************************************************************
 Abbruch der Verifikationsalgorithmen
 ----------------------------------------------------------------------
 verification algorithms cancelled by user
 **********************************************************************}
procedure TForm1.ButtonAbbruchClick(Sender: TObject);
begin
  Cancelled := true;
end;

{**********************************************************************
 durch Probieren geeignete Koeffizienten fr die Berechnung des Tages
 im Jahr aus einem gegebenen Datum ermitteln
 ----------------------------------------------------------------------
 determine suitable coefficients for computing number of day within year
 from given date
 **********************************************************************}
procedure TForm1.ButtonDate2DayNumberClick(Sender: TObject);
var a, b, c, m: integer;
    Hit: boolean;
begin

  Cancelled := false;
  ListViewKoeff.Items.Clear;
  ButtonCancelCoeff.Enabled := true;
  CheckBoxSJ.Enabled := false;
  Screen.Cursor := crHourGlass;

  for a := 100 to 700 do begin
    Application.ProcessMessages;
    if Cancelled then
      break;
    for b := -50 to 50 do
      for c := 1 to 100 do begin
        { probiere aktuelles Koeffiziententripel a, b, c
          ----------------------------------------------
          try current coefficients a, b, c }
        for m := 2 to 12 do begin
          if CheckBoxSJ.Checked then
            Hit := f (a, b, c, m) = sly [m]
          else
            Hit := f (a, b, c, m) = s [m];
          if not Hit then
            break;
        end;
        if Hit then
          with ListViewKoeff.Items.Add do begin
            Caption := IntToStr (a);
            SubItems.Add (IntToStr (b));
            SubItems.Add (IntToStr (c));
          end;

      end;
  end;

  ButtonCancelCoeff.Enabled := false;
  CheckBoxSJ.Enabled := true;
  Screen.Cursor := crDefault;

end;

{**********************************************************************
 durch Probieren geeignete Koeffizienten fr die Berechnung des Datums
 anhand des Tages im Jahr ermitteln
 ----------------------------------------------------------------------
 determine suitable coefficients for computing date from number of day
 within year
 **********************************************************************}
procedure TForm1.ButtonDayNumber2DateClick(Sender: TObject);
var a, b, c, i: integer;
    Hit: boolean;

begin

  ListViewKoeff.Items.Clear;
  Cancelled := false;
  ButtonCancelCoeff.Enabled := true;
  CheckBoxSJ.Enabled := false;
  Screen.Cursor := crHourGlass;

  for a := 1 to 100 do begin
    Application.ProcessMessages;
    if Cancelled then
      break;
    for b := 1 to 600 do
      for c := 1 to 700 do begin
        { probiere aktuelles Koeffiziententripel a, b, c
          ----------------------------------------------
          try current coefficients a, b, c }
        for i := 2 to 23 do begin
          if CheckBoxSJ.Checked then
            Hit := f (a, b, c, nly [i]) = m [i]
          else
            Hit := f (a, b, c, n [i]) = m [i];
          if not Hit then
            break;
        end;
        if Hit then
          with ListViewKoeff.Items.Add do begin
            Caption := IntToStr (a);
            SubItems.Add (IntToStr (b));
            SubItems.Add (IntToStr (c));
          end;

      end;
  end;

  ButtonCancelCoeff.Enabled := false;
  CheckBoxSJ.Enabled := true;
  Screen.Cursor := crDefault;

end;

{**********************************************************************
 Algorithmus zur Bestimmung der Julianischen Tagesnummer verifizieren.
 Dies erfolgt durch den Nachweis, da die beiden verwendeten Funktionen
 bijektiv sind. Dazu wird herangezogen der

 Satz:
 =====
 Eine Funktion f: A -> B ist genau dann Bijektion, wenn es eine Funktion
 g: B -> A gibt, fr die g o f = id (A) und f o g = id (B) gelten.
 **********************************************************************}
procedure TForm1.ButtonStartJDNClick(Sender: TObject);

var Jahr, Monat, Tag: integer;
    EndJahr, EndMonat, EndTag: integer;
    n, nStrich: integer;
    Datum: string;
    OK: boolean;

begin

  OK := true;
  Cancelled := false;
  ButtonStartJDN.Enabled := false;
  ButtonCancelJDN.Enabled := true;
  Screen.Cursor := crHourGlass;

  EndJahr  := StrToInt (MaskEditJDNJahr.Text);
  EndMonat := StrToInt (MaskEditJDNMonat.Text);
  EndTag   := StrToInt (MaskEditJDNTag.Text);

  { Verifikation Teil 1: aus allen JDN Datum ermitteln
    und dann Umkehrfunktion anwenden }
  for n := JulianDayNumber ('15821015') to
    JulianDayNumber (YMDtoYYYYMMDD (EndJahr, EndMonat, EndTag))
  do begin
    if n mod 120 = 0 then begin
      Application.ProcessMessages;
      if Cancelled then
        break;
    end;
    Datum := JDNtoYYYYMMDD (n);
    nStrich := JulianDayNumber (Datum);
    if nStrich <> n then begin
      { Fehler }
      ShowMessage ('Fehler in Teil 1 bei JDN = ' + IntToStr (n));
      OK := false;
      break;
    end;
  end;

  { Verifikation Teil 2: aus allen gltigen Datumsangaben JDN ermitteln
    und dann Umkehrfunktion anwenden }
  Jahr := 1582;
  Monat := 10;
  Tag := 15;

  repeat

    if Monat mod 4 = 0 then begin
      Application.ProcessMessages;
      if Cancelled then
        break;
    end;

    { JDN aus Datum berechnen }
    n := JulianDayNumber (YMDtoYYYYMMDD (Jahr, Monat, Tag));
    { Umkehrfunktion: Datum aus JDN berechnen }
    Datum := JDNtoYYYYMMDD (n);
    if (Jahr <> ExtractYear (Datum)) or (Monat <> ExtractMonth (Datum)) or
      (Tag <> ExtractDay (Datum))
    then begin
      { soben ermitteltes Datum stimmt nicht mit ursprnglichem Datum berein }
      ShowMessage ('Fehler in Teil 2 bei JDN = ' + IntToStr (n));
      OK := false;
      break;
    end;

    { einen Tag weiter }
    inc (Tag);
    if Tag > DaysOfMonth (Jahr, Monat) then begin
      inc (Monat);
      Tag := 1;
    end;
    if Monat > 12 then begin
      inc (Jahr);
      Monat := 1;
    end;

  { Schleife endet, wenn letztes zu verifizierendes Datum verarbeitet }
  until (Jahr = EndJahr) and (Monat = EndMonat) and (Tag = EndTag);

  ButtonStartJDN.Enabled := true;
  ButtonCancelJDN.Enabled := false;
  Screen.Cursor := crDefault;

  if OK and not Cancelled then
    ShowMessage ('Algorithmus ok!');

end;

{**********************************************************************
 Algorithmus zur Bestimmung des Tages im Jahr verifizieren.
 Dies erfolgt durch den Nachweis, da die beiden verwendeten Funktionen
 bijektiv sind. Dazu wird herangezogen der

 Satz:
 =====
 Eine Funktion f: A -> B ist genau dann Bijektion, wenn es eine Funktion
 g: B -> A gibt, fr die g o f = id (A) und f o g = id (B) gelten.
 **********************************************************************}
procedure TForm1.ButtonStartDayNumberClick(Sender: TObject);
var Jahr, Monat, Tag: integer;
    n, nStrich: integer;
    Datum: string;
    OK: boolean;

begin

  OK := true;
  Cancelled := false;
  ButtonStartDayNumber.Enabled := false;
  Screen.Cursor := crHourGlass;

  if RadioButton1999.Checked then
    Jahr := 1999
  else
    Jahr := 2000;
  Monat := 1;
  Tag := 1;

  { Verifikation Teil 1: aus allen fortlaufenden Nummern n = 1,...365/366
    Datum ermitteln und dann Umkehrfunktion anwenden }
  for n := 1 to 365 + IsLeapYear (Jahr) do begin
    Datum := DayNumberToYYYYMMDD (Jahr, n);
    nStrich := DayNumber (Datum);
    if nStrich <> n then begin
      { Fehler }
      ShowMessage ('Fehler in Teil 1 bei TIJ = ' + IntToStr (n));
      OK := false;
      break;
    end;
  end;

  { Verifikation Teil 2: aus allen gltigen Datumsangaben TIJ ermitteln
    und dann Umkehrfunktion anwenden }

  repeat

    { TIJ aus Datum berechnen }
    n := DayNumber (YMDtoYYYYMMDD (Jahr, Monat, Tag));
    { Umkehrfunktion: Datum aus TIJ berechnen }
    Datum := DayNumberToYYYYMMDD (Jahr, n);
    if (Jahr <> ExtractYear (Datum)) or (Monat <> ExtractMonth (Datum)) or
      (Tag <> ExtractDay (Datum))
    then begin
      { soben ermitteltes Datum stimmt nicht mit ursprnglichem Datum berein }
      ShowMessage ('Fehler in Teil 2 bei TIJ = ' + IntToStr (n));
      OK := false;
      break;
    end;

    { einen Tag weiter }
    inc (Tag);
    if Tag > DaysOfMonth (Jahr, Monat) then begin
      inc (Monat);
      Tag := 1;
    end;

  { Schleife endet, NACHDEM 31.12. erreicht }
  until (Monat = 13) and (Tag = 1);

  ButtonStartDayNumber.Enabled := true;
  Screen.Cursor := crDefault;

  if OK and not Cancelled then
    ShowMessage ('Algorithmus ok!');

end;

{**********************************************************************
 eingegebenes Datum berprfen und formatieren
 **********************************************************************}
procedure TForm1.ButtonScanClick(Sender: TObject);
var YYYYMMDD: string;
    Year: integer;

begin
  LabelYYYYMMDD.Caption := 'YYYYMMDD: ';
  LabelDatumFormatiert.Caption := 'Datum formatiert: ';
  LabelKW.Caption := 'KW (DIN): ';

  if not ScanDateString (EditDatum.Text, YYYYMMDD) then
    MessageBeep (0)
  else begin
    LabelYYYYMMDD.Caption := 'YYYYMMDD: ' + YYYYMMDD;
    Year := ExtractYear (YYYYMMDD);
    LabelKW.Caption := 'KW (DIN): ' + IntToStr (WeekNumberISO (YYYYMMDD, Year)) +
      '/' + IntToStr (Year);
    LabelDatumFormatiert.Caption := 'Datum formatiert: ' +
      FormatDate (YYYYMMDD, true);
  end;
end;

{**********************************************************************
 Ostersonntag im Jahr berechnen
 ----------------------------------------------------------------------
 compute Easter Sunday in year
 **********************************************************************}
procedure TForm1.ButtonOsternClick(Sender: TObject);
begin
  LabelOstersonntag.Caption := 'Ostersonntag: ' +
    FormatDate (EasterSunday (StrToInt (MaskEditOsterJahr.Text)), true);
end;

{**********************************************************************
 Komponente zur Datumseingabe dynamisch erzeugen
 (ermglicht Compilierung auch ohne vorherige Installation der Komponente)
 ----------------------------------------------------------------------
 create component TCMDateEdt dynamically (compiling is possible prior
 to installing component)
 **********************************************************************}
procedure TForm1.FormCreate(Sender: TObject);
begin
  CMDateEdt1 := TCMDateEdt.Create (Self);
  with CMDateEdt1 do begin
    Parent := GroupBoxTCMDateEdt;
    Left := 10;
    Top := 20;
    Width := 75;
  end;
end;

{**********************************************************************
 **********************************************************************}
procedure TForm1.ButtonValidateClick(Sender: TObject);
begin
  LabelYYYYMMDDvalid.Caption := 'YYYYMMDD: ' + CMDateEdt1.DateYYYYMMDD;
end;

{**********************************************************************
 **********************************************************************}
procedure TForm1.ButtonKWClick(Sender: TObject);
var Year, YearWeek1, Week: integer;
    January1st: string;
begin

  ListViewKW.Items.Clear;
  
  for Year := StrToInt (MaskEditJahrVon.Text) to StrToInt (MaskEditJahrBis.Text) do
    with ListViewKW.Items.Add do begin
      Caption := IntToStr (Year);
      January1st := YMDtoYYYYMMDD (Year, 1, 1);

      { Wochentag des 01.01. }
      SubItems.Add (ShortDayName (WeekDayISO (January1st)));

      { 01.01. in KW }
      Week := WeekNumberISO (January1st, YearWeek1);
      SubItems.Add (Format ('%2.2d/%4.4d', [Week, YearWeek1]));

      { Beginn KW 1 }
      SubItems.Add (FormatDate (BeginWeek1ISO (Year), true));
    end;

end;

end.
